
#include <malloc.h>
#include "list.h"

void                     listInit(List * list, CompFn compFn, word keyOffset, Boolean ownsItems);
void                     listDeinit(List * list);
word                     listNitems(List * list);
void                     listIterateInit(List * list, ListIterationType iterationType);
void                    *listIterate(List * list, ListNode ** node);
void                     listIterateDeinit(List * list);
void                    *listPeekHead(List * list, ListNode ** node);
void                    *listPeekTail(List * list, ListNode ** node);
void                     listAddHead(List * list, void *item);
void                     listAddTail(List * list, void *item);
void                     listDel(List * list, ListNode * node);
void                    *listSearch(List * list, ListNode ** node, void *key);

void
listInit(List * list, CompFn compFn, word keyOffset, Boolean ownsItems)
{
  list->nItems = 0;
  list->head = list->tail = NULL;
  list->compFn = compFn;
  list->keyOffset = keyOffset;
  list->flags = 0;
  if (ownsItems)
    list->flags |= LIST_OWNS_ITEMS;
}

void
listDeinit(List * list)
{
  ListNode                *p = list->head;

  while (p) {
    ListNode                *p2 = p->next;

    if (list->flags & LIST_OWNS_ITEMS)
      free(p->item);
    free(p);
    p = p2;
  }
  list->head = NULL;
}

word
listNitems(List * list)
{
  return list->nItems;
}

void
listIterateInit(List * list, ListIterationType iterationType)
{
  list->flags |= LIST_ITERATING;
  if (iterationType == HEAD_TO_TAIL) {
    list->flags &= ~LIST_ITERATE_TAIL_TO_HEAD;
    list->curNode = list->head;
  } else {
    list->flags |= LIST_ITERATE_TAIL_TO_HEAD;
    list->curNode = list->tail;
  }
}

void *
listIterate(List * list, ListNode ** node)
{
  if (list->curNode) {
    void                    *res;

    res = list->curNode->item;
    if (node)
      *node = list->curNode;
    if (list->flags & LIST_ITERATE_TAIL_TO_HEAD)
      list->curNode = list->curNode->prev;
    else
      list->curNode = list->curNode->next;
    return res;
  } else {
    listIterateDeinit(list);
    if (node)
      *node = NULL;
    return NULL;
  }
}

void
listIterateDeinit(List * list)
{
  list->flags &= ~LIST_ITERATING;
}

void *
listPeekHead(List * list, ListNode ** node)
{
  if (node)
    *node = list->head;
  if (list->head)
    return list->head->item;
  else
    return NULL;
}

void *
listPeekTail(List * list, ListNode ** node)
{
  if (node)
    *node = list->tail;
  if (list->tail)
    return list->tail->item;
  else
    return NULL;
}

void
listAddHead(List * list, void *item)
{
  ListNode                *listNode = (ListNode *) malloc(sizeof(ListNode));

  listNode->prev = NULL;
  listNode->next = list->head;
  listNode->item = item;
  if (list->head)
    list->head->prev = listNode;
  else
    list->tail = listNode;
  list->head = listNode;
  list->nItems++;
}

void
listAddTail(List * list, void *item)
{
  ListNode                *listNode = (ListNode *) malloc(sizeof(ListNode));

  listNode->prev = list->tail;
  listNode->next = NULL;
  listNode->item = item;
  if (list->tail)
    list->tail->next = listNode;
  else
    list->head = listNode;
  list->tail = listNode;
  list->nItems++;
}

void
listDel(List * list, ListNode * node)
{
  if ((list->flags & LIST_ITERATING) && (node == list->curNode)) {
    if (list->flags & LIST_ITERATE_TAIL_TO_HEAD)
      list->curNode = node->prev;
    else
      list->curNode = node->next;
  }
  if (node->prev)
    node->prev->next = node->next;
  else
    list->head = node->next;
  if (node->next)
    node->next->prev = node->prev;
  else
    list->tail = node->prev;
  if (list->flags & LIST_OWNS_ITEMS)
    free(node->item);
  free(node);
  list->nItems--;
}

void *
listSearch(List * list, ListNode ** node, void *key)
{
  CompFn                   compFn = list->compFn;
  ListNode                *p = list->head;

  while (p) {
    if (!compFn((byte *) p->item + list->keyOffset, key)) {
      if (node)
	*node = p;
      list->curNode = p;
      return p->item;
    }
    p = p->next;
  }
  if (node)
    *node = NULL;
  list->curNode = NULL;
  return NULL;
}

void *
listSearchNext(List *list, ListNode **node, void *key)
{
  CompFn                   compFn = list->compFn;
  ListNode                *p = list->curNode;

  if (p) p = p->next;
  while (p) {
    if (!compFn((byte *)p->item + list->keyOffset, key)) {
      if (node)
	*node = p;
      list->curNode = p;
      return p->item;
    }
    p = p->next;
  }
  if (node)
    *node = NULL;
  list->curNode = NULL;
  return NULL;
}
